td.smallinfo {
    font-size: 8px;
}

body {
    font-family: "Quicksand", "Helvetica Neue", Helvetica, Arial;
}

h1, h2, h3 {
    font-weight: bold;
}

h1 {
    font-size: 20px;
    color: #bb4400;
    background-color: #ffeecc;
    padding-top: 3px;
    padding-bottom: 2px;
    border-bottom: 1px solid #ccc;
    border-top: 1px solid #cce;
    padding-left: 10px;
}

h2 {
    font-size: 16px;
    text-transform: uppercase;
}

table {
    margin: 8px;
    border-collapse: collapse;
    border-bottom: 1px solid #ddd;
}

table#objects-container {
    margin: 0px;
    border: none;
}

table#objects-container tr td {
    vertical-align: top;
}

@keyframes rotation {
  from {
    transform: rotate(0deg);
  }
  to {
    transform: rotate(359deg);
  }
}

div.progress-spinner {
    display: inline-block;
    width: 1em;
    height: 1em;
    background-image: url('/static/icons/tangos-50x50.png');
    background-size: cover;
    background-repeat: no-repeat;
    animation: rotation 2s infinite;
    animation-timing-function: linear;
}
tr.tangos-data:nth-child(odd) {
    background-color: #eeeef3;
}

tr.tangos-data:nth-child(even):hover {
    background-color: #eaeaea;
}

tr.tangos-data:nth-child(odd):hover {
    background-color: #dadae1;
}


td, th {
    padding: 2px;
    padding-left: 5px;
    padding-right: 5px;
}

td {
    font-family: 'Lucida Console', 'Lucida Sans Typewriter', monaco, 'Bitstream Vera Sans Mono', monospace;
    font-size: small;
    color: #222;
}

td.editable, th.editable {
    color: #bb4400;
}

a {
    color: #bb4400;
    text-decoration: none;
}

a:hover {
    color: #cc5511;
}



th {
    border-bottom: 1px solid #ddd;
    border-top: 1px solid #ddd;
}

#breadcrumbs-top {
    padding-top: 10px;
    padding-bottom: 4px;
    font-size: 12px;
}

#breadcrumbs-bottom {
    border-top: 1px solid #dde;
    clear: both;
    padding-top: 20px;
    font-size: 12px;
}

#tree-json {
    display: none;
}

#imgbox {
    overflow: scroll;
    height: 100%;
}

#imgbox::-webkit-scrollbar {
    -webkit-appearance: none;
}

#imgbox::-webkit-scrollbar:vertical {
    width: 11px;
}

#imgbox::-webkit-scrollbar:horizontal {
    height: 11px;
}

#imgbox::-webkit-scrollbar-thumb {
    border-radius: 8px;
    border: 2px solid white;
    background-color: rgba(0, 0, 0, .5);
}

#imgbox::-webkit-scrollbar-track {
    background-color: #fff;
    border-radius: 8px;
}

#imgbox {

}

img.plotimg {
    float: right;
}

span.large {
    font-size:30px;
}

span.small_big {
    font-size:12px;
}

span.step_available a {

}

span.step_unavailable a {
    color: #f00;
}

#imgbox_container {
    position: fixed;
    float: right;
    right: 0px;
    top: 0px;
    bottom: 0px;
    min-width: 150px;
    background-color: #ffffff;
    z-index: 200;
    max-width: 80vw;
    padding: 1em;
    box-shadow: -10px 0 10px white;
    display: flex;
    flex-direction: column;
}


#imgspace {


}

#navigation {
    width: 100%;
    background-color: rgba(223, 223, 223, 0.9);
    border-bottom: 1px solid gray;
    padding-left: 10px;
    z-index: 100;
}

#flash-message{
    color: #500;
}

.hidden {
    display: none;
}

code  {
	font-size: 10px;
}

div.popup_parent {
    top:0;
    right:0;
    bottom:0;
    left:0;
    background-color: #fff;
    opacity: 0.5;
    z-index: 2;
    visibility: hidden;
    position: fixed;
}

div.popup_box {

    left: 20px;
    top: 100px;
    width: 300px;
    padding: 10px;
    font-size: 12px;
    background-color: #dde;
    z-index: 3;
    position: fixed;
    visibility: hidden;
}

pre {
    font-family: 'Lucida Console', 'Lucida Sans Typewriter', monaco, 'Bitstream Vera Sans Mono', monospace;
    white-space: pre-wrap;
    font-size: small;
}

.load_table_failed::before {
    content: " ";
    display: inline-block;
    width: 1.28em;
    height: 1em;
    margin-right: 0.2em;
    background-image: url('/static/icons/error.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

.mono {
    font-family: 'Lucida Console', 'Lucida Sans Typewriter', monaco, 'Bitstream Vera Sans Mono', monospace;
    font-size: small;
    color: #222;
}

td.plotcontrols {
    display: flex;
    flex-direction: row;
    justify-content: space-between;
}

th.plotcontrols div.leftbuttons, th.plotcontrols div.rightbuttons {
    display: inline;
}

div.leftbuttons, div.rightbuttons {
    white-space: nowrap;
}

div.rightbuttons {
    margin-left: 1em;
}


/* Merger tree */

g.node {
    font-family: Verdana, Helvetica;
    font-size: 12px;
    font-weight: bold;
}

circle.node-dot-standard {
    fill: #bb4400;
    stroke: #bb4400;
    stroke-width: 1px;
}

circle.node-dot-sub {
    fill: #bb4400;
    stroke: bb4400;
    stroke-width: 1px;
}

circle.node-dot-selected {
    fill: #e26000;
    stroke:black;
    stroke-width: 3px;
}

.phantom {
    stroke-dasharray: 2,2;
    fill-opacity: 0.5;
}

circle.node-dot-continuation {
    fill: white;
    stroke: #bbb;
    stroke-width: 1px;
}

path.link {
    fill: none;
    stroke: #e2e2e2;
}

text.info {
  font-family: Verdana, Helvetica;
  font-size: 12px;
  fill: gray;
}

rect.info {
  fill: white;
  stroke: gray;
}

table#halos-container tr td {
    vertical-align: top;
}

table.data-table tr th {
    min-width: 30px;
}

div.popup-controls {
    background-color: #eeeeee;
    font-size: large;
    padding: 3px;
    cursor: pointer;
}

div#properties-area {
    overflow: hidden;
    width: 50%;
}

ul.ui-autocomplete {
    max-height: 200px;
    overflow-y: scroll;
    overflow-x: hidden;
    font-family: 'Lucida Console', 'Lucida Sans Typewriter', monaco, 'Bitstream Vera Sans Mono', monospace;
    font-size: small;
}



input[type=checkbox].filter-checkbox {
    display: none;
}

label.filter-checkbox {
    cursor: pointer;
}

label.filter-checkbox::before {
    content: " ";
    display: inline-block;
    width: 1.27em;
    height: 1em;
    background-image: url('/static/icons/filter.svg');
    background-size: cover;
    background-repeat: no-repeat;
    cursor: pointer;
}

input[type=checkbox]:checked + label.filter-checkbox::before {
    background-image: url('/static/icons/filter_active.svg');
}

a.sort.asc::before {
    content: " ";
    display: inline-block;
    width: 1.27em;
    height: 1em;
    background-image: url('/static/icons/sort_up.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

a.sort.asc.active::before {
    background-image: url('/static/icons/sort_up_active.svg');
}

a.sort.desc::before {
    content: " ";
    display: inline-block;
    width: 1.27em;
    height: 1em;
    background-image: url('/static/icons/sort_down.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

a.sort.desc.active::before {
    background-image: url('/static/icons/sort_down_active.svg');
}

a.delete::before {
    content: " ";
    display: inline-block;
    width: 1.011em;
    height: 1em;
    background-image: url('/static/icons/delete.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

a.download-link {
    white-space: nowrap;
}

label.x-plot-radio::before {
    content: " ";
    display: inline-block;
    width: 1.1348em; /* 1.4494em; */
    height: 1em;
    background-image: url('/static/icons/plot.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

/*
label.y-plot-radio::before {
    content: " ";
    display: inline-block;
    width: 1.359em;
    height: 1em;
    background-image: url('/static/icons/plot_y.svg');
    background-size: cover;
    background-repeat: no-repeat;
}
 */

label.plot-radio::before {
    content: " ";
    display: inline-block;
    width: 1.1348em;
    height: 1em;
    background-image: url('/static/icons/plot.svg');
    background-size: cover;
    background-repeat: no-repeat;
}

input {
    accent-color: #BB4400;
}

#imgbox_container .ui-resizable-w {
    border-right: 2px solid rgba(187, 68, 0, 0.2);
    cursor: ew-resize;
    z-index: 0;
    flex-grow: 1;
}
